//
// chronometer on the LCD display
// 
// external counter interrupts 4 times per second
// displays shows a tumbling wheel;  (hex) counter increments every second
//

#include "cMIPS.h"

extern int _counter_val;

int conv1(int);

#define FALSE (0==1)
#define TRUE  !FALSE

#define QUARTER 12500000

int main(void) {
  int i, j, k;
  volatile int old_val;
  int sec, min, hour;

  LCDinit();

  LCDtopLine();

#if 1
  LCDput(' ');
  LCDput('H');
  LCDput('e');
  LCDput('l');
  LCDput('l');
  LCDput('o');
  LCDput(' ');
  LCDput('w');
  LCDput('o');
  LCDput('r');
  LCDput('l');
  LCDput('d');
  LCDput('!');
#else
  LCDprint( print_sp() ); // for debugging only
#endif

  LCDbotLine();

  _counter_val = 0;   // variable to accumulate number of interrupts

  old_val = _counter_val;

  enableInterr();
  startCounter(QUARTER,TRUE); // counter will interrupt after 1/4 second

  sec = min = hour = 0;

  i = 0;

  while (TRUE) {
  
    while (old_val == _counter_val) { // busy wait for interrupt
      delay_us(i);   // that's really stupid but we have nothing else to do
    }

    old_val = _counter_val;

    switch(i) {
    case 0:
      j = 0x09; break;
    case 1:
      j = 0x0b; break;
    case 2:
      j = 0x0c; break;
    default:
      j = 0xa;
      i = -1;
      sec += 1;
      break;
    };
    i += 1;
    DSP7SEGput(0, 0, 0, 0, (i<<2)); // blink red led
 
    LCDgotoxy(1, 2);
    LCDput(j);

    if (sec == 60) {
      min += 1;
      sec = 0;
    }

    if (min == 60) {
      hour += 1;
      min = 0;
    }

    LCDgotoxy(4, 2);

#define conv(a) ((a<10)?((a)+0x30):((a)+('a'-10)))

    k = (hour>>4) & 0x0f;
    LCDput( conv(k) );
    k = hour & 0x0f;
    LCDput( conv(k) );

    LCDput(':');

    k = (min>>4) & 0x0f;
    LCDput( conv(k) );
    k = min & 0x0f;
    LCDput( conv(k) );

    LCDput(':');

    k = (sec>>4) & 0x0f;
    LCDput( conv(k) );
    k = sec & 0x0f;
    LCDput( conv(k) );

  }

  return 0;

}



int conv1(int c) {
  int i;

    switch(c) {
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
      i = c + 0x30; ; break;
    case 10:
      i = 'a'; break;
    case 11:
      i = 'b'; break;
    case 12:
      i = 'c'; break;
    case 13:
      i = 'd'; break;
    case 14:
      i = 'e'; break;
    case 15:
      i = 'f'; break;
    default:
      i = '*';
    }

    return(i);
}
